@@ -1,4 +1,7 @@
 import requests
+from typing import Dict
+import urllib3
+urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
 
 
 class REST:
@@ -18,7 +21,7 @@ class REST:
         :param ssl: Usage of SSL certificate, defaults to False
         :type ssl: bool, optional
         """
-        self.url = url + "/v1/api/"
+        self.url = f"{url}/v1/api/"
         self.ssl = ssl
         self.id = self.get_accounts()[0]["accountId"]
 
@@ -28,7 +31,7 @@ class REST:
         :return: list of account info
         :rtype: list
         """
-        response = requests.get(self.url + "portfolio/accounts", verify=self.ssl)
+        response = requests.get(f"{self.url}portfolio/accounts", verify=self.ssl)
         return response.json()
 
     def switch_account(self, accountId: str) -> dict:
@@ -40,8 +43,9 @@ class REST:
         :rtype: dict
         """
         response = requests.post(
-            self.url + "iserver/account", json={"acctId": accountId}, verify=self.ssl
+            f"{self.url}iserver/account", json={"acctId": accountId}, verify=self.ssl
         )
+
         self.id = accountId
         return response.json()
 
@@ -52,9 +56,9 @@ class REST:
         :rtype: float
         """
         response = requests.get(
-            self.url + "portfolio/" + self.id + "/ledger",
-            verify=self.ssl,
+            f"{self.url}portfolio/{self.id}/ledger", verify=self.ssl
         )
+
         return response.json()["USD"]["cashbalance"]
 
     def get_netvalue(self) -> float:
@@ -64,24 +68,65 @@ class REST:
         :rtype: float
         """
         response = requests.get(
-            self.url + "portfolio/" + self.id + "/ledger",
-            verify=self.ssl,
+            f"{self.url}portfolio/{self.id}/ledger", verify=self.ssl
         )
+
         return response.json()["USD"]["netliquidationvalue"]
 
-    def get_conid(self, symbol: str) -> int:
+    def get_conid(
+        self,
+        symbol: str,
+        instrument_filters: Dict = None,
+        contract_filters: Dict = {"isUS": True},
+    ) -> int:
         """Returns contract id of the given stock instrument
 
         :param symbol: Symbol of the stock instrument
         :type symbol: str
+        :param instrument_filters: Key-value pair of filters to use on the returned instrument data, e.g) {"name": "ISHARES NATIONAL MUNI BOND E", "assetClass": "STK"}
+        :type instrument_filters: Dict, optional
+        :param contract_filters: Key-value pair of filters to use on the returned contract data, e.g) {"isUS": True, "exchange": "ARCA"}
+        :type contract_filters: Dict, optional
         :return: contract id
         :rtype: int
         """
         query = {"symbols": symbol}
         response = requests.get(
-            self.url + "trsrv/stocks", params=query, verify=self.ssl
+            f"{self.url}trsrv/stocks", params=query, verify=self.ssl
         )
+
         dic = response.json()
+
+        if instrument_filters or contract_filters:
+
+            def filter_instrument(instrument: Dict) -> bool:
+                def apply_filters(x: Dict, filters: Dict) -> list:
+                    positives = list(
+                        filter(
+                            lambda x: x,
+                            [x.get(key) == val for key, val in filters.items()],
+                        )
+                    )
+                    return len(positives) == len(filters)
+
+                if instrument_filters:
+                    valid = apply_filters(instrument, instrument_filters)
+
+                    if not valid:
+                        return False
+
+                if contract_filters:
+                    instrument["contracts"] = list(
+                        filter(
+                            lambda x: apply_filters(x, contract_filters),
+                            instrument["contracts"],
+                        )
+                    )
+
+                return len(instrument["contracts"]) > 0
+
+            dic[symbol] = list(filter(filter_instrument, dic[symbol]))
+
         return dic[symbol][0]["contracts"][0]["conid"]
 
     def get_portfolio(self) -> dict:
@@ -91,13 +136,11 @@ class REST:
         :rtype: dict
         """
         response = requests.get(
-            self.url + "portfolio/" + self.id + "/positions/0",
-            verify=self.ssl,
+            f"{self.url}portfolio/{self.id}/positions/0", verify=self.ssl
         )
-        dic = {}
-        for item in response.json():
-            dic.update({item["contractDesc"]: item["position"]})
-        dic.update({"USD": self.get_cash()})  # Cash balance
+
+        dic = {item["contractDesc"]: item["position"] for item in response.json()}
+        dic["USD"] = self.get_cash()
         return dic
 
     def reply_yes(self, id: str) -> dict:
@@ -112,10 +155,9 @@ class REST:
 
         answer = {"confirmed": True}
         response = requests.post(
-            self.url + "iserver/reply/" + id,
-            json=answer,
-            verify=self.ssl,
+            f"{self.url}iserver/reply/{id}", json=answer, verify=self.ssl
         )
+
         return response.json()[0]
 
     def _reply_all_yes(self, response, reply_yes_to_all: bool) -> dict:
@@ -141,10 +183,11 @@ class REST:
         :rtype: dict
         """
         response = requests.post(
-            self.url + "iserver/account/" + self.id + "/orders",
+            f"{self.url}iserver/account/{self.id}/orders",
             json={"orders": list_of_orders},
             verify=self.ssl,
         )
+
         return self._reply_all_yes(response, reply_yes)
 
     def get_order(self, orderId: str) -> dict:
@@ -156,11 +199,12 @@ class REST:
         :rtype: dict
         """
         response = requests.get(
-            self.url + "iserver/account/order/status/" + str(orderId), verify=self.ssl
+            f"{self.url}iserver/account/order/status/{orderId}", verify=self.ssl
         )
+
         return response.json()
 
-    def get_live_orders(self, filters: list = []) -> dict:
+    def get_live_orders(self, filters: list = None) -> dict:
         """Returns list of live orders
 
         :param filters: List of filters for the returning response. Available items -- "inactive" "pending_submit" "pre_submitted" "submitted" "filled" "pending_cancel" "cancelled" "warn_state" "sort_by_time", defaults to []
@@ -168,11 +212,14 @@ class REST:
         :return: list of live orders
         :rtype: dict
         """
+        if filters is None:
+            filters = []
         response = requests.get(
-            self.url + "iserver/account/orders",
+            f"{self.url}iserver/account/orders",
             params={"filters": filters},
             verify=self.ssl,
         )
+
         return response.json()
 
     def cancel_order(self, orderId: str) -> dict:
@@ -184,9 +231,9 @@ class REST:
         :rtype: dict
         """
         response = requests.delete(
-            self.url + "iserver/account/" + self.id + "/order/" + str(orderId),
-            verify=self.ssl,
+            f"{self.url}iserver/account/{self.id}/order/{orderId}", verify=self.ssl
         )
+
         return response.json()
 
     def modify_order(
@@ -208,10 +255,11 @@ class REST:
         ), "Input parameters (orderId or order) are missing"
 
         response = requests.post(
-            self.url + "iserver/account/" + self.id + "/order/" + str(orderId),
+            f"{self.url}iserver/account/{self.id}/order/{orderId}",
             json=order,
             verify=self.ssl,
         )
+
         return self._reply_all_yes(response, reply_yes)
 
     def ping_server(self) -> dict:
@@ -220,7 +268,7 @@ class REST:
         :return: Response from the server
         :rtype: dict
         """
-        response = requests.post(self.url + "tickle", verify=self.ssl)
+        response = requests.post(f"{self.url}tickle", verify=self.ssl)
         return response.json()
 
     def get_auth_status(self) -> dict:
@@ -229,17 +277,17 @@ class REST:
         :return: Status dictionary
         :rtype: dict
         """
-        response = requests.post(self.url + "iserver/auth/status", verify=self.ssl)
+        response = requests.post(f"{self.url}iserver/auth/status", verify=self.ssl)
         return response.json()
 
     def re_authenticate(self) -> None:
         """Attempts to re-authenticate when authentication is lost"""
-        response = requests.post(self.url + "iserver/reauthenticate", verify=self.ssl)
+        requests.post(f"{self.url}iserver/reauthenticate", verify=self.ssl)
         print("Reauthenticating ...")
 
     def log_out(self) -> None:
         """Log out from the gateway session"""
-        response = requests.post(self.url + "logout", verify=self.ssl)
+        requests.post(f"{self.url}logout", verify=self.ssl)
 
     def get_bars(
         self,
@@ -274,8 +322,9 @@ class REST:
             "outsideRth": outsideRth,
         }
         response = requests.get(
-            self.url + "iserver/marketdata/history", params=query, verify=self.ssl
+            f"{self.url}iserver/marketdata/history", params=query, verify=self.ssl
         )
+
         return response.json()
 
     def get_fut_conids(self, symbol: str) -> list:
@@ -288,8 +337,9 @@ class REST:
         """
         query = {"symbols": symbol}
         response = requests.get(
-            self.url + "trsrv/futures", params=query, verify=self.ssl
+            f"{self.url}trsrv/futures", params=query, verify=self.ssl
         )
+
         return response.json()[symbol]
 
 
@@ -314,7 +364,9 @@ if __name__ == "__main__":
     # print(api.get_portfolio())
     # print(api.re_authenticate())
     # print(api.get_auth_status())
-    print(api.get_live_orders())
+    # print(api.get_live_orders())
     # print(api.get_bars("TSLA"))
     # print(api.get_conid("AAPL"))
+    # print(api.get_conid("MUB", contract_filters={"isUS": True, "exchange": "ARCA"}))
+    # print(api.get_conid("MUB", instrument_filters={"name": "ISHARES NATIONAL MUNI BOND E", "assetClass": "STK"}))
     # print(api.cancel_order(2027388848))
